/* File: SRC\CHECKSUM.C */

/**************************************************************************
The Internet Extender ver. 1.0 beta is a software based Network Address 
Translating Router designed to allow for multiple machines on a LAN to 
simultaneously connect through a single IP address to the Internet.
Copyright (C) 1999  Jason Sarich, Dr. Dan Lewis, Santa Clara University

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
**************************************************************************/

#include "napt.h"

void Fix_IP_Checksum(in_Header *ip)
	{
	/* ----------------------------------------- */
	/* Calculate the checksum for the IP header. */
	/* ----------------------------------------- */
	ip->checksum = 0 ;
	ip->checksum = ~Checksum(ip, IP_HDR_BYTES(ip)) ;
	}

void Fix_Prot_Checksum(in_Header *ip)
	{
	register byte *prot ;
	tcp_PseudoHeader ph ;
	unsigned length ;
	word *prot_cksm ;

	/* ----------------------------------- */
	/* "length" is the length of IP header */
	/* ----------------------------------- */
	length = IP_HDR_BYTES(ip) ;

	/* ------------------------------------ */
	/* "prot" is a pointer to the beginning */
	/* of the protocol (TCP or UDP) header. */
	/* ------------------------------------ */
	prot = (byte *) ip + length ;

	/* --------------------------------------------------- */
	/* "length" becomes length of data following IP header */
	/* --------------------------------------------------- */
	length = Rev16(ip->length) - length ;

	/* ---------------------------------------- */
	/* "prot_cksm" is a pointer to the checksum */
	/* in the protocol (TCP or UDP) header.     */
	/* ---------------------------------------- */
	switch (ip->proto)
		{
		case TCP_PROTO:
			prot_cksm = &((tcp_Header *) prot)->checksum ;
			break ;

		case UDP_PROTO:
			prot_cksm = &((udp_Header *) prot)->checksum ;
			if (!*prot_cksm) return ;
			break ;
		}

	/* ----------------------------------------------- */
	/* Initialize protocol checksum before calculating */
	/* the checksum for the pseudo header.		   */
	/* ----------------------------------------------- */
	*prot_cksm = 0 ;

	/* ------------------------------------------------- */
	/* Calculate the checksum for the pseudo header.     */
	/* (This is used to calculate the protocol checksum) */
	/* ------------------------------------------------- */
	ph.src = ip->source ;
	ph.dst = ip->destination ;
	ph.mbz = 0 ;
	ph.protocol = ip->proto ;
	ph.length = Rev16(length) ;
	ph.checksum = Checksum(prot, length) ;

	/* ---------------------------------------------- */
	/* Calculate the checksum for the protocol header */
	/* ---------------------------------------------- */
	*prot_cksm = ~Checksum(&ph, sizeof(ph)) ;
	}

void Fix_ICMP_Checksum(in_Header *ip)
	{
	register ICMP_UNUSED *icmp ;
	unsigned ip_hdr_len ;

	ip_hdr_len = IP_HDR_BYTES(ip) ;
	icmp = (ICMP_UNUSED *) ((byte *) ip + ip_hdr_len) ;
	icmp->checksum = 0 ;
	icmp->checksum = ~Checksum(icmp, Rev16(ip->length) - ip_hdr_len) ;
	}
