/* File: SRC\DISPLAY.C */

/**************************************************************************
The Internet Extender ver. 1.0 beta is a software based Network Address 
Translating Router designed to allow for multiple machines on a LAN to 
simultaneously connect through a single IP address to the Internet.
Copyright (C) 1999  Jason Sarich, Dr. Dan Lewis, Santa Clara University

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
**************************************************************************/

#include "napt.h"

#define	ATT_NORMAL		0
#define	ATT_INVERSE		1
#define	ATT_INDICATOR		2
#define	ATT_PROTOCOL		3
#define	ATT_LCL_ADRS		4
#define	ATT_REM_ADRS		5
#define	ATT_MAPPING		6
#define	ATT_PKT_COUNT		7
#define	ATT_EXPIRE		8
#define	ATT_ERR_TXT		9

#define	ATTRIBUTES		10

#define	SCREEN_ROWS		25
#define	SCREEN_COLS		80

#define	FIRST_ROW		0
#define	FIRST_COL		0

#define	LAST_ROW		(SCREEN_ROWS - 1)
#define	LAST_COL		(SCREEN_COLS - 1)

#define	MSG_SECONDS		60
#define	MSG_COLUMN		4
#define	MAX_SHRINK		(SCREEN_ROWS - 4)

#define	CHAR_BYTES		2
#define	ROW_BYTES		(SCREEN_COLS * CHAR_BYTES)
#define	SCREEN_SIZE		(SCREEN_ROWS * ROW_BYTES)
#define	ROWCOL(r, c)		((r)*ROW_BYTES + (c)*CHAR_BYTES)

PRIVATE byte			Attribute(int) ;
PRIVATE void			Display_Hex(char *, void *, int) ;
PRIVATE int			Display_Offset(int) ;
PRIVATE void			Display_Mapping(MAPPING far *) ;
PRIVATE int			Display_Row(int) ;
PRIVATE void			Erase_Row(int, int) ;
PRIVATE void			Erase_Rows(int, int) ;
PRIVATE void			Move_Rows(int, int, int) ;
PRIVATE int			PutStrAt(char *, int, int, int) ;
PRIVATE void			Scroll_Indicators(void) ;
PRIVATE void			Shrink_Window(void) ;

PRIVATE SEMA_INFO *		display_sema = NULL ;
PRIVATE BOOL			color = TRUE ;
PRIVATE byte far *		video_buffer ;
PRIVATE int			footer = LAST_ROW ;
PRIVATE longword		msg_timer = 0L ;

void Init_Display(void)
	{
	word video_seg ;

	MtCSemaCreate(&display_sema) ;
	color = (biosequip() & 0x30) != 0x30 ;
	video_seg = color ? 0xB800 : 0xB000 ;
	video_buffer = MK_FP(video_seg, 0) ;
	clrscrn() ;
	poscurs(LAST_ROW, LAST_COL) ;
	cursoff() ;
	atexit(curson) ;
	}

void Display(void)
	{
	Erase_Row(0, ATT_INVERSE) ;
	PutStrAt("Proto", ATT_INVERSE, 0, 1) ;
	PutStrAt("Remote IP:Port", ATT_INVERSE, 0, 12) ;
	PutStrAt("Local IP:Port", ATT_INVERSE, 0, 36) ;
	PutStrAt("Mapping", ATT_INVERSE, 0, 52) ;
	PutStrAt("Rcvd", ATT_INVERSE, 0, 61) ;
	PutStrAt("Sent", ATT_INVERSE, 0, 67) ;
	PutStrAt("Expire", ATT_INVERSE, 0, 73) ;

	Erase_Row(footer, ATT_INVERSE) ;
	PutStrAt(NOTICE, ATT_INVERSE, footer, 1) ;

	MtCSetPri(NULL, MTC_LOW) ;
	Show_Percent() ;
	Visit_Mappings(Display_Mapping) ;
	}

PRIVATE void Scroll_Indicators(void)
	{
	byte far *top = video_buffer + ROWCOL(FIRST_ROW, LAST_COL) ;
	byte far *btm = video_buffer + ROWCOL(footer, LAST_COL) ;

	if (Display_Offset(0))
		{
		top[0] = '' ;
		top[1] = Attribute(ATT_INDICATOR) ;
		}
	else
		{
		top[0] = ' ' ;
		top[1] = Attribute(ATT_INVERSE) ;
		}

	if (Mappings(0) - Display_Offset(0) >= footer)
		{
		btm[0] = '' ;
		btm[1] = Attribute(ATT_INDICATOR) ;
		}
	else
		{
		btm[0] = ' ' ;
		btm[1] = Attribute(ATT_INVERSE) ;
		}
	}

BOOL Scroll_Key(word scan_ascii)
	{
	int delta ;

	switch (scan_ascii)
		{
		case 0x4800: /* Up Arrow */
			delta = -1 ;
			break ;
		case 0x4900: /* Page Up */
			delta = -(SCREEN_ROWS - 3) ;
			break ;
		case 0x5000: /* Down Arrow */
			delta = 1 ;
			break ;
		case 0x5100: /* Page Down */
			delta = SCREEN_ROWS - 3 ;
			break ;
		default:
			return FALSE ;
		}

	(void) Display_Offset(delta) ;
	return TRUE ;
	}

PRIVATE int Display_Offset(int delta)
	{
	static int offset = 0 ;
	offset += delta ;
	if (offset < 0) offset = 0 ;
	if (Mappings(0) && offset >= Mappings(0))
		{
		offset = Mappings(0) - 1 ;
		}
	return offset ;
	}

PRIVATE int Display_Row(int mapping)
	{
	int row = 1 + mapping - Display_Offset(0) ;
	return (0 < row && row < footer) ? row : 0 ;
	}

PRIVATE byte Attribute(int code)
	{
	static byte attributes[2][ATTRIBUTES] =
		{
		{ /* MONO */
		/* ATT_NORMAL	*/	0x07,
		/* ATT_INVERSE	*/	0x70,
		/* ATT_INDICATOR*/	0xF8,
		/* ATT_PROTOCOL	*/	0x07,
		/* ATT_LCL_ADRS	*/	0x07,
		/* ATT_REM_ADRS	*/	0x07,
		/* ATT_MAPPING	*/	0x07,
		/* ATT_PKT_COUNT*/	0x07,
		/* ATT_EXPIRE	*/	0x07,
		/* ATT_ERR_TXT	*/	0x0F
		},
		{/* COLOR */
		/* ATT_NORMAL	*/	0x07,
		/* ATT_INVERSE	*/	0x70,
		/* ATT_INDICATOR*/	0xF8,
		/* ATT_PROTOCOL	*/	0x0E,
		/* ATT_LCL_ADRS	*/	0x02,
		/* ATT_REM_ADRS	*/	0x03,
		/* ATT_MAPPING	*/	0x0A,
		/* ATT_PKT_COUNT*/	0x07,
		/* ATT_EXPIRE	*/	0x04,
		/* ATT_ERR_TXT	*/	0x0C
		}
		} ;

	return attributes[color][code] ;
	}

void Update_Display(void)
	{
	static longword cycle = 0L ;

	if (Check_Timeout(cycle))
		{
		longword now ;

		MtCSemaReset(display_sema) ;
		now = Set_Timeout(0) ;
		cycle = now + (Set_Timeout(10) - now + 15) / 30 ;
		}
	}

PRIVATE void Display_Mapping(MAPPING far *m)
	{
	char bfr[81], *proto ;
	unsigned minutes, hours ;
	static int mapping = 0 ;
	longword now, seconds ;
	int row, col ;

	if (msg_timer && Check_Timeout(msg_timer))
		{
		Restore_Window() ;
		msg_timer = 0L ;
		}

	if (!m)
		{
		MtCSemaWait(display_sema, FALSE) ;
		MtCSemaSet(display_sema) ;
		if (row = Display_Row(mapping))
			{
			Erase_Rows(row, footer - 1) ;
			}
		mapping = 0 ;
		Scroll_Indicators() ;
		return ;
		}

	if (m->status & STATUS_SERVER) return ;

	row = Display_Row(mapping++) ;
	if (!row) return ;
	col = 2 ;

	switch (m->c.protocol)
		{
		case TCP_PROTO:		proto = "TCP  " ;	break ;
		case UDP_PROTO:		proto = "UDP  " ;	break ;
		case ICMP_PROTO:	proto = "ICMP " ;	break ;
		default:
			sprintf(proto = bfr, "%02Xh  ", m->c.protocol) ;
			break ;
		}
	col = PutStrAt(proto, ATT_PROTOCOL, row, col) ;

	sprintf(bfr, "%16s", IP2Ascii(m->c.remote_IP)) ;
	col = PutStrAt(bfr, ATT_REM_ADRS, row, col) ;
	if (m->c.protocol == ICMP_PROTO) strcpy(bfr, "       ") ;
	else sprintf(bfr, ":%-6u", m->c.remote_port) ;
	col = PutStrAt(bfr, ATT_REM_ADRS, row, col) ;

	sprintf(bfr, "%16s", IP2Ascii(m->c.local_IP)) ;
	col = PutStrAt(bfr, ATT_LCL_ADRS, row, col) ;
	if (m->c.protocol == ICMP_PROTO) strcpy(bfr, "       ") ;
	else sprintf(bfr, ":%-6u", m->c.local_port) ;
	col = PutStrAt(bfr, ATT_LCL_ADRS, row, col) ;

	sprintf(bfr, "%5u ", m->mapped_port) ;
	col = PutStrAt(bfr, ATT_MAPPING, row, col) ;

	sprintf(bfr, "%6lu", m->pkt_rcvd) ;
	col = PutStrAt(bfr, ATT_PKT_COUNT, row, col) ;
	sprintf(bfr, "%6lu", m->pkt_sent) ;
	col = PutStrAt(bfr, ATT_PKT_COUNT, row, col) ;

	now = Set_Timeout(0) ;
	if (now >= m->time_out) seconds = 0L ;
	else seconds = (64L * (m->time_out - now) + (1165L/2L)) / 1165L ;
	hours = (unsigned) (seconds / 3600L) ;	seconds %= 3600L ;
	minutes = (unsigned) (seconds / 60L) ;	seconds %= 60L ;

	sprintf(bfr, " %02u:%02u:%02lu", hours, minutes, seconds) ;
	PutStrAt(bfr, ATT_EXPIRE, row, col) ;
	}

void Show_Percent(void)
	{
	char bfr[10] ;

	sprintf(bfr, "[%3u%c]", Percent_Used(), '%') ;
	PutStrAt(bfr, ATT_INVERSE, footer, LAST_COL - strlen(bfr)) ;
	}

PRIVATE void Erase_Rows(int first, int last)
	{
	static word blank = 0x0720 ;
	unsigned blanks ;
	void far *fp ;

	fp = video_buffer + ROWCOL(first, 0) ;
	blanks = SCREEN_COLS * (last - first + 1) ;
	farrepmem(fp, (void far *) &blank, sizeof(blank), blanks) ;
	}

PRIVATE void Erase_Row(int row, int attb)
	{
	static byte blank[2] = {' '} ;
	void far *fp = video_buffer + ROWCOL(row, 0) ;
	blank[1] = (byte) Attribute(attb) ;
	farrepmem(fp, (void far *) blank, sizeof(blank), SCREEN_COLS) ;
	}

PRIVATE int PutStrAt(char *str, int attb, int row, int col)
	{
	byte far *fp ;
	int c ;

	fp = video_buffer + ROWCOL(row, col) ;
	attb = Attribute(attb) ;
	while (c = *str++)
		{
		if (col == SCREEN_COLS) break ;
		*fp++ = (byte) c ;
		*fp++ = (byte) attb ;
		col++ ;
		}

	return col ;
	}

PRIVATE void Display_Hex(char *label, void *data, int len)
	{
	char bfr[81] ;
	char *bp, *p ;
	int i, col ;

	Shrink_Window() ;
	col = PutStrAt(label, ATT_NORMAL, LAST_ROW, MSG_COLUMN) ;
	p = (char *) data ; bp = bfr ;
	for (i = 0; i < len; i++)
		{
		sprintf(bp, " %02X", *p++) ;
		bp += 3 ;
		}
	*bp = '\0' ;
	PutStrAt(bfr, ATT_LCL_ADRS, LAST_ROW, col) ;
	}

void Warning(CONNECTION *c, in_Header *ip, char *err, ...)
	{
	static unsigned prev_row, prev_col, repeat_count ;
	static CONNECTION *old_c, prev_c = {0} ;
	static char *prev_err = NULL ;
	char bfr[81], *prot, *phdr, *data ;
	va_list argptr ;
	unsigned col ;
	int dlen ;

	if (msg_timer != 0L && err == prev_err && c == old_c &&
	    (!c || !memcmp(c, &prev_c, sizeof(CONNECTION))))
		{
		sprintf(bfr, " (%u)", ++repeat_count) ;
		PutStrAt(bfr, ATT_NORMAL, prev_row, prev_col) ;
		msg_timer = Set_Timeout(MSG_SECONDS) ;
		return ;
		}

	prev_err = err ; prev_c = *c ; old_c = c ; repeat_count = 1 ;
	msg_timer = Set_Timeout(MSG_SECONDS) ;

	if (footer != LAST_ROW) Shrink_Window() ;

	va_start(argptr, err) ;
	vsprintf(bfr, err, argptr) ;
	Shrink_Window() ;
	col = PutStrAt(" Error Message: ", ATT_NORMAL, LAST_ROW, MSG_COLUMN) ;
	prev_col = PutStrAt(bfr, ATT_ERR_TXT, LAST_ROW, col) ;
	prev_row = LAST_ROW ;

	if (!c) return ;

	phdr = ((char *) ip) + IP_HDR_BYTES(ip) ;

	switch (c->protocol)
		{
		case TCP_PROTO:
			prot = "TCP" ;
			data = phdr +
				(DATAOFFSET(((tcp_Header *) phdr)) << 2) ;
			break ;

		case UDP_PROTO:
			prot = "UDP" ;
			data = phdr + sizeof(udp_Header) ;
			break ;

		case ICMP_PROTO:
			prot = "ICMP" ;
			data = NULL ;
			break ;

		default:
			sprintf(prot = bfr, "%02Xh", c->protocol) ;
			data = NULL ;
			break ;
		}

	if (data)
		{
		dlen = Rev16(ip->length) - (data - (char *) ip) ;
		if (dlen <= 0) data = NULL ;
		else if (dlen > 20) dlen = 20 ;
		}

	Shrink_Window() ; prev_row-- ;
	col = PutStrAt("      Protocol: ", ATT_NORMAL, LAST_ROW, MSG_COLUMN) ;
	PutStrAt(prot, ATT_PROTOCOL, LAST_ROW, col) ;

	Shrink_Window() ; prev_row-- ;
	col = PutStrAt("Remote Address: ", ATT_NORMAL, LAST_ROW, MSG_COLUMN) ;
	col = PutStrAt(IP2Ascii(c->remote_IP), ATT_REM_ADRS, LAST_ROW, col) ;
	if (c->protocol != ICMP_PROTO && c->remote_port)
		{
		sprintf(bfr, ":%u", c->remote_port) ;
		PutStrAt(bfr, ATT_REM_ADRS, LAST_ROW, col) ;
		}

	Shrink_Window() ; prev_row-- ;
	col = PutStrAt(" Local Address: ", ATT_NORMAL, LAST_ROW, MSG_COLUMN) ;
	col = PutStrAt(IP2Ascii(c->local_IP), ATT_LCL_ADRS, LAST_ROW, col) ;
	if (c->protocol != ICMP_PROTO && c->local_port)
		{
		sprintf(bfr, ":%u", c->local_port) ;
		PutStrAt(bfr, ATT_LCL_ADRS, LAST_ROW, col) ;
		}

	if (!ip) return ;

	Display_Hex("     IP Header:", ip, 20) ; prev_row-- ;

	if (!FRAG_OFFSET(ip))
		{
		sprintf(bfr, "   %4s Header:", prot) ;
		Display_Hex(bfr, phdr, 20) ; prev_row-- ;
		}

	if (data)
		{
		sprintf(bfr, "     %4s Data:", prot) ;
		Display_Hex(bfr, data, dlen) ; prev_row-- ;
		}
	}

PRIVATE void Move_Rows(int rows, int fm, int to)
	{
	byte far *src = video_buffer + ROWCOL(fm, 0) ;
	byte far *dst = video_buffer + ROWCOL(to, 0) ;
	farmemmove(dst, src, rows * ROW_BYTES) ;
	}

BOOL Restore_Window(void)
	{
	if (footer == LAST_ROW) return FALSE ;
	Move_Rows(1, footer, LAST_ROW) ;
	Erase_Rows(footer, LAST_ROW - 1) ;
	footer = LAST_ROW ;
	return TRUE ;
	}

PRIVATE void Shrink_Window(void)
	{
	int cur_rows ;

	cur_rows = LAST_ROW - footer ;
	if (cur_rows + 1 <= MAX_SHRINK)
		{
		Move_Rows(cur_rows + 1, footer, footer - 1) ;
		footer-- ;
		}
	else Move_Rows(cur_rows - 1, footer + 2, footer + 1) ;
	Erase_Rows(LAST_ROW, LAST_ROW) ;
	}
