/*
	File: FILTER.C
*/

#include "..\include\wattcp.h"
#include "..\include\filter.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>

static FILTER *Decode_Filter(char *bfr) ;

FILTER *Filter_Match(longword adrs, word port, FILTER *block, FILTER *allow)
	{
	FILTER *f, *blocked ;

	for (f = block; f; f = f->next)
		{
		/* Check for a matching IP address */
		if ((adrs ^ f->adrs) & f->mask) continue ;

		/* Does the associated port match? */
		if (!f->port || port == f->port) break ;
		}

	if (!f) return NULL ;

	blocked = f ;
	for (f = allow; f; f = f->next)
		{
		/* Check for a matching IP address */
		if ((adrs ^ f->adrs) & f->mask) continue ;

		/* Does the associated port match? */
		if (!f->port || port == f->port) break ;
		}

	return f ? NULL : blocked ;
	}

FILTER *Add2Filter(FILTER *filter, char *bfr)
	{
	FILTER *f, *new ;

	f = Decode_Filter(bfr) ;
	if (new = Filter_Match(f->adrs, f->port, filter, NULL)) return new ;
	if (new = (FILTER *) malloc(sizeof(FILTER)))
		{
		memcpy(new, f, sizeof(FILTER)) ;
		new->next = filter ;
		}

	return new ;
	}

static FILTER *Decode_Filter(char *bfr)
	{
	static FILTER filter ;

	/* ------------------------------------------------------------ */
	/*	format: <ip_adrs>[:<port>][,<mask>[,<frwd>[:<port>]]]	*/
	/* ------------------------------------------------------------ */

	filter.next = NULL ;
	filter.adrs = aton(bfr) ;
	filter.mask = 0xFFFFFFFFL ;
	filter.port = 0x0000 ;
	filter.frwd = 0x00000000L ;
	filter.fprt = 0x0000 ;

	while (strchr("0123456789.", *bfr)) bfr++ ;
	if (*bfr == ':') filter.port = (word) atoi(++bfr) ;

	bfr = strchr(bfr, ',') ;
	if (!bfr || !*++bfr) return &filter ;

	if (*bfr != ',') filter.mask = aton(bfr) ;

	bfr = strchr(bfr, ',') ;
	if (!bfr || !*++bfr) return &filter ;

	filter.frwd = aton(bfr) ;
	bfr = strchr(bfr, ':') ;
	if (!bfr || !*++bfr) return &filter ;

	filter.fprt = (word) atoi(bfr) ;
	if (filter.fprt != 0 && filter.port == 0) return NULL ;

	return &filter ;
	}

