/*
	File: LICENSE.C
*/

/*
#define	DEBUG
*/

#define	STATIC_LICENSE		1
#define	FLOATING_LICENSE	2
#define	LICENSE_TYPE		STATIC_LICENSE

#include "..\include\wattcp.h"
#include "..\include\elib.h"
#include <conio.h>
#include <memory.h>
#include <stdlib.h>
#include <bios.h>

#define	CAPACITY	500	/* Maximum number of licensees	*/
#define	LIFETIME	10	/* Seconds before entry expires */

typedef struct {		/* Stolen from PCRECV.C	*/
    longword	 buf_sig;
    longword	 buf_hisip;
    word         buf_hisport;
    int		 buf_len;
    byte	 buf_data[ 1500 ];
} recv_buf;

#define	ONE_UDP_BFR	sizeof(recv_buf) /* Datagram look-ahead capacity */
#define	UDP_BFR_SIZE	(4 * ONE_UDP_BFR)

#define	LICENSE_PORT	1946	/* Dan Lewis's birth year       */
#define	BROADCAST_IP	0xFFFFFFFFL

#define	HELLO		0xF0	/* It's alive!	*/
#define	WORKING		0xE1	/* Still alive!	*/
#define	GOODBYE		0xD2	/* Call again!	*/
#define	KILL		0xC3	/* Over limit!	*/

#undef	NULL
#define	NULL		((void *) 0)

#define	BOOL		int
#define	PRIVATE		static
#define	FALSE		0
#define	TRUE		1

typedef struct LICENSE_DATAGRAM
	{
	byte	cmd ;
	long	ip ;
	int	node ;
	} LICENSE_DATAGRAM ;

typedef struct LICENSE_STATUS
	{
	long	ip ;		/* IP of running node; 0 if vacant	*/
	long	expires ;	/* Time that this entry expires.	*/
	} LICENSE_STATUS ;

PRIVATE void Remove_Entry(long) ;
PRIVATE void Post_Entry(long, BOOL) ;
PRIVATE void License_Exit(void);
PRIVATE void Broadcast(byte, long) ;

#ifdef	DEBUG
PRIVATE void Status_Message(char *, long) ;
PRIVATE void Display_List(void) ;
#endif

extern udp_Socket *udp_socket(void) ;

PRIVATE LICENSE_STATUS licenses[CAPACITY] ;
PRIVATE int high_water = 0 ;
PRIVATE int licensees = 0 ;
PRIVATE udp_Socket *s ;
PRIVATE unsigned port ;
PRIVATE long add_me_timer = 0L ;
PRIVATE int this_node = 0 ;

int License_Init(unsigned this_pgm_id, int limit, int node)
	{
	static byte udp_bfr[UDP_BFR_SIZE] ;

	if ((licensees = limit) >= 999) return ;

	this_node = node ;
	port = LICENSE_PORT + this_pgm_id ;
	if (licensees > CAPACITY) licensees = CAPACITY ;
	memset(licenses, 0, sizeof(licenses)) ;

	s = udp_socket() ;
	if (!s)
		{
		puts("Insufficient memory for UDP License socket") ;
		exit(255) ;
		}

	udp_open(s, port, BROADCAST_IP, port, NULL) ;
	if (sock_recv_init(s, udp_bfr, sizeof(udp_bfr)))
		{
		puts("Can't enable floating license buffer") ;
		exit(255) ;
		}

	sock_mode((sock_type *) s, UDP_MODE_NOCHK) ;
	atexit(License_Exit) ;
	Broadcast(HELLO, 0L) ;
	return port ;
	}

BOOL Unlicensed(void)	/* Returns TRUE if told to DIE! */
	{
	LICENSE_DATAGRAM action ;
	LICENSE_STATUS *l ;
	int i, size ;

	if (licensees >= 999) return FALSE ;

	/* --------------------------- */
	/* (1) Remind others I'm alive */
	/* --------------------------- */
	if (chk_timeout(add_me_timer)) Broadcast(WORKING, 0L) ;

	/* ------------------------------ */
	/* (2) Remove any expired entries */
	/* ------------------------------ */
	l = licenses ;
	for (i = 0; i < high_water; i++, l++)
		{
		if (chk_timeout(l->expires))
			{
			l->ip = 0L ;
			if (i == high_water - 1) high_water-- ;
			}
		}

	/* ----------------------------- */
	/* (3) Process pending datagrams */
	/* ----------------------------- */
	while (size = sock_recv(s, (char *) &action, sizeof(action), 0))
		{
#ifdef	DEBUG
		char ip[20] ;
		inet_ntoa(ip, action.ip) ;
#endif
		if (size != sizeof(action)) continue ;

		/* ---------------------------------------- */
		/* Protect against modified DOOR.SYS files! */
		/* ---------------------------------------- */
		if (action.node == this_node) return TRUE ;

		/* ------------------------- */
		/* Decide what to do with it */
		/* ------------------------- */
		switch (action.cmd)
			{
			case HELLO:
				Post_Entry(action.ip, TRUE) ;
#ifdef	DEBUG
				Status_Message("Received HELLO from %s\n",
					ip) ;
#endif
				break ;

			case WORKING:
				Post_Entry(action.ip, FALSE) ;
#ifdef	DEBUG
				Status_Message("Received WORKING from %s\n",
					ip) ;
#endif
				break ;

			case KILL:
#ifdef	DEBUG
				Status_Message("Received KILL for %s\n", ip) ;
#endif
				if (action.ip == gethostid()) return TRUE ;
				break ;

			case GOODBYE:
				Remove_Entry(action.ip) ;
#ifdef	DEBUG
				Display_List() ;
				Status_Message("Received GOODBYE from %s\n",
					ip) ;
#endif
				break ;

#ifdef	DEBUG
			default:
				Status_Message("Received UNKNOWN from %s\n",
					ip) ;
#endif
			}
		}

	return FALSE ;
	}

PRIVATE void License_Exit(void)
	{
	Broadcast(GOODBYE, 0L) ;
	sock_close((sock_type *) s) ;
	}

PRIVATE void Post_Entry(long ip, BOOL new)
	{
	LICENSE_STATUS *l, *vacant ;
	int i, count ;

	vacant = NULL ;
	count = 1 ;
	l = licenses ;
	for (i = 0; i < high_water; i++, l++)
		{
		if (l->ip == ip)
			{
			l->expires = set_timeout(LIFETIME) ;
			return ;
			}

		if (!l->ip && !vacant)
			{
			vacant = l ;
			}

		else count++ ;
		}

	if (count >= licensees)
		{
		Broadcast(KILL, ip) ;
		return ;
		}

	if (!vacant)
		{
		if (high_water >= licensees)
			{
			if (new) Broadcast(KILL, ip) ;
			return ;
			}

		vacant = &licenses[high_water++] ;
		}

	vacant->ip = ip ;
	vacant->expires = set_timeout(LIFETIME) ;
#ifdef	DEBUG
	Display_List() ;
#endif
	}

PRIVATE void Remove_Entry(long ip)
	{
	LICENSE_STATUS *l ;
	int i ;

	l = licenses ;
	for (i = 0; i < high_water; i++, l++)
		{
		if (l->ip == ip)
			{
			l->ip = 0L ;
			if (i == high_water - 1) high_water-- ;
			break ;
			}
		}
	}

PRIVATE void Broadcast(byte cmd, long ip)
	{
	LICENSE_DATAGRAM action ;
#ifdef	DEBUG
	char *str ;

	switch (cmd)
		{
		case HELLO:	str = "HELLO" ;		break ;
		case WORKING:	str = "WORKING" ;	break ;
		case GOODBYE:	str = "GOODBYE" ;	break ;
		case KILL:	str = "KILL" ;		break ;
		default:	str = "UNKNOWN" ;	break ;
		}
	Status_Message("Broadcast(%s)\n", (long) str) ;
#endif

	action.cmd = (byte) cmd ;
	action.ip = ip ? ip : gethostid() ;
	action.node = this_node ;
	sock_fastwrite((sock_type *) s, (byte *) &action, sizeof(action)) ;
	add_me_timer = set_timeout(LIFETIME / 2) ;
	}

#ifdef	DEBUG
extern void sock_init(void) ;
extern void tcp_tick(void *) ;

#define	MAXLINE 24 
PRIVATE int display_line = 0 ;
PRIVATE long display_timer = 0L ;

PRIVATE void Status_Message(char *str, long ip)
	{
	printf(str, ip) ;
	if (++display_line > MAXLINE) Display_List() ;
	}

PRIVATE void Display_List(void)
	{
	char ip[20] ;
	LICENSE_STATUS *l ;
	int i, n ;

	clrscrn() ;
	printf("Currently running nodes:\n\n") ;
	inet_ntoa(ip, gethostid()) ;
	printf("\tNode #%d: %s\n", n = 1, ip) ;
	display_line = 4 ;

	l = licenses ;
	for (i = 0; i < high_water; i++, l++)
		{
		if (l->ip)
			{
			inet_ntoa(ip, l->ip) ;
			printf("\tNode #%d: %s\n", ++n, ip) ;
			display_line++ ;
			}
		}
	printf("\n") ;

	display_timer = set_timeout(2 * LIFETIME) ;
	}

main(int argc, char *argv[])
	{
	if (argc != 3)
		{
		puts("\n\tUsage: license <max_nodes> <this_node#>") ;
		exit(255) ;
		}

	sock_init() ;

	License_Init(1, atoi(argv[1]), atoi(argv[2])) ;
	Display_List() ;
	for (;;)
		{
		if (Unlicensed()) exit(0) ;

		tcp_tick(NULL) ;

		if (bioskey(1))
			{
			(void) bioskey(0) ;
			break ;
			}
		}
	}
#endif
