#include <wattcp.h>
#include <malloc.h>
#include <mem.h>

#define RECV_USED    0xF7E3D2B1L
#define RECV_UNUSED  0xE6D2C1AFL

typedef struct
	{
	longword	recv_sig;
	byte		*recv_bufs;
	word		recv_bufnum;	/* for udp */
	word		recv_len;	/* for tcp */
	} RECV_DATA;

typedef struct
	{
	longword	buf_sig;
	longword	buf_hisip;
	word		buf_hisport;
	int		buf_len;
	byte		buf_data[ 1500 ];
	} RECV_BUF;


/*
 *  recvd - gets upcalled when data arrives
 */
int _recvdaemon(udp_Socket *s, byte FAR *data, int len, tcp_PseudoHeader *ph, void FAR *h)
	{
	RECV_DATA FAR *r = NULL ;
	RECV_BUF *p;
	tcp_Socket *t;
	udp_Header FAR *uh;
	int i;

	/* check for length problem */
	if (!len || len >= 1500) return 0 ;

	switch (s->ip_type)
		{
		case UDP_PROTO:
			r = (RECV_DATA FAR *) s->rdata ;
                        uh = (udp_Header FAR *) h;
			if (r->recv_sig != RECV_USED)
				{
				outs("ERROR: udp recv data conflict") ;
				return 0 ;
				}

			p = (RECV_BUF *) r->recv_bufs ;

			/* find a buffer */
			for (i = 0; i < r->recv_bufnum; ++i , ++p)
				{
#ifdef	__POWERC
				if (p->buf_sig == RECV_UNUSED)
					{
					/* take this one */
					p->buf_sig = RECV_USED ;
					p->buf_len = len ;
					p->buf_hisip = ph->src ;
					p->buf_hisport = uh->srcPort ;
					_MEMCPY(p->buf_data, data, len) ;
					return 0 ;
					}

				if (p->buf_sig != RECV_USED)
					{
					outs("ERROR: sock_recv_init data err");
					return 0 ;
					}
#else
				switch (p->buf_sig)
					{
					case RECV_USED:
						break;

					case RECV_UNUSED:
						/* take this one */
						p->buf_sig = RECV_USED ;
						p->buf_len = len ;
						p->buf_hisip = ph->src ;
						p->buf_hisport = uh->srcPort ;
						_MEMCPY(p->buf_data,data,len);
						return 0 ;

					default	   :
						outs("ERROR: sock_recv_init data err");
						return 0 ;
					}
#endif
				}
			return 0 ;

		case TCP_PROTO :
			t= (tcp_Socket *) s ;
			r = (RECV_DATA FAR *) t->rdata ;
			if (r->recv_sig != RECV_USED)
				{
				outs("ERROR: udp recv data conflict");
				return 0 ;
				}

			/* stick it on the end if you can */
			i = t->maxrdatalen - t->rdatalen;
			if (i > 1)
				{
				/* we can accept some of this */
				if (len > i) len = i;
				_MEMCPY(&r->recv_bufs[s->rdatalen],
				   data, len);
				s->rdatalen += len ;
				return len ;
				}

			return 0 ;	/* didn't take none */
		}

	return 0 ;
	}

void sock_recv_init(udp_Socket *s, void *space, word len)
	{
	tcp_Socket *t ;
	RECV_DATA FAR *r = NULL ;
	RECV_BUF *p ;
	int i ;

	/* clear data area */
	memset(p = space, 0, len);

	s->dataHandler = _recvdaemon;

	/* clear table */
	_MEMSET(r = (RECV_DATA *) s->rddata, 0, tcp_MaxRxSize + 1) ;
	r->recv_sig = RECV_USED ;
	r->recv_bufs = (byte *) p ;
	r->recv_bufnum = len / sizeof(RECV_BUF) ;
	r->recv_len = len ;

	if (s->ip_type == UDP_PROTO)
		{
		for (i = 0 ; i < r->recv_bufnum ; ++i, ++p)
			{
			p->buf_sig = RECV_UNUSED ;
			}
		}
	}

int sock_recv_from(udp_Socket *s, long *hisip, word *hisport,
		   char *buffer, int len, word flags)
	{
	tcp_Socket *t;
	RECV_DATA FAR *r = NULL ;
	RECV_BUF *p;
	int i;

	r = (RECV_DATA FAR *) s->rdata ;

	if (r->recv_sig != RECV_USED) return -1 ;

	switch (s->ip_type)
		{
		case UDP_PROTO:
			p = (RECV_BUF *) r->recv_bufs ;

			/* find a buffer */
			for (i = 0; i < r->recv_bufnum; ++i , ++p)
				{
#ifdef	__POWERC
				if (p->buf_sig == RECV_USED)
					{
					p->buf_sig = RECV_USED ;
					if (len > p->buf_len)
						{
						len = p->buf_len ;
						}

					memcpy(buffer, p->buf_data, len) ;
					if (hisip) *hisip = p->buf_hisip ;
					if (hisport)
						{
						*hisport = p->buf_hisport;
						}
					p->buf_sig = RECV_UNUSED ;
					return len ;
					}

				if (p->buf_sig != RECV_UNUSED)
					{
					outs("ERROR: sock_recv_init data err");
					return 0 ;
					}
#else
				switch (p->buf_sig)
					{
					case RECV_UNUSED:
						break;

					case RECV_USED:
						p->buf_sig = RECV_USED ;
						if (len > p->buf_len)
							{
							len = p->buf_len ;
							}
						memcpy(buffer, p->buf_data, 
							len) ;
						if (hisip)
							{
							*hisip = p->buf_hisip;
							}
						if (hisport)
							{
							*hisport = p->buf_hisport;
							}
						p->buf_sig = RECV_UNUSED ;
						return len ;
					default:
						outs("ERROR: sock_recv_init data err");
						return 0 ;
					}
#endif
				}
			return 0 ;

		case TCP_PROTO:
			t = (tcp_Socket *) s ;
			if (len > t->rdatalen) len = t->rdatalen ;
			if (len) memcpy(buffer, r->recv_bufs, len) ;
			return len ;
		}

	return 0 ;
	}

int sock_recv(udp_Socket *s, char *buffer, int len, word flags)
	{
	tcp_Socket *t;
	RECV_DATA FAR *r = NULL ;
	RECV_BUF *p;
	int i;

	r = (RECV_DATA FAR *) s->rdata;

	if (r->recv_sig != RECV_USED) return -1 ;

	switch (s->ip_type)
		{
		case UDP_PROTO:
			p = (RECV_BUF *) r->recv_bufs;

			/* find a buffer */
			for (i = 0; i < r->recv_bufnum; ++i , ++p)
				{
#ifdef	__POWERC
				if (p->buf_sig == RECV_USED)
					{
					p->buf_sig = RECV_USED ;
					if (len > p->buf_len)
						{
						len = p->buf_len;
						}
					memcpy(buffer, p->buf_data, len) ;
					p->buf_sig = RECV_UNUSED ;
					return len ;
					}

				if (p->buf_sig != RECV_UNUSED)
					{
					outs("ERROR: sock_recv_init data err");
					return 0 ;
					}
#else
				switch (p->buf_sig)
					{
					case RECV_UNUSED:
						break;

					case RECV_USED:
						p->buf_sig = RECV_USED ;
						if (len > p->buf_len)
							{
							len = p->buf_len ;
							}
						memcpy(buffer, p->buf_data, 
							len) ;
						p->buf_sig = RECV_UNUSED ;
						return len ;

					default:
						outs("ERROR: sock_recv_init data err");
						return 0 ;
					}
#endif
				}
			return 0 ;

		case TCP_PROTO:
			t = (tcp_Socket*) s;
			if ( len > t->rdatalen ) len = t->rdatalen ;
			if (len) memcpy(buffer, r->recv_bufs, len) ;
			return len ;
		}

	return 0 ;
	}

