#include <wattcp.h>
#include <dos.h>

#define	PRIVATE		static
#define	PUBLIC		

#define PD_PPP		0x0006

#define INT_FIRST	0x0060
#define INT_LAST	0x0080

#define PD_DRIVER_INFO	0x01FF
#define PD_ACCESS 	0x0200
#define PD_RELEASE	0x0300
#define PD_SEND		0x0400
#define PD_GET_ADDRESS	0x0600

#define CARRY 		0x0001	     /* carry bit in flags register */

extern void _ppp_init(void far *, int, int) ;
extern void _ppp_isr(void) ;

#define	PPP_MAXBFRS	4
#define	PPP_BFRSIZE	1518

PRIVATE char far	_pkt_line[] = "PKT DRVR" ;
PRIVATE byte FAR	_ppp_pkts[PPP_MAXBFRS][PPP_BFRSIZE + 2] ;
PRIVATE word		_ppp_handle ;
PRIVATE word		_ppp_interrupt ;

PRIVATE void		ppp_fini(void) ;
PRIVATE int		is_ppp_driver(int) ;

PUBLIC void ppp_init(void)
	{
	static word _ppp_ip_type = 0x0008 ;	/* intelled value */
	static void far *fptr_ppp_ip_type = (void far *) &_ppp_ip_type ;
	static void far *fptr_ppp_isr = (void far *) _ppp_isr ;
	struct REGPACK regs ;
	int ivec ;

	_ppp_init(&_ppp_pkts[0][0], PPP_MAXBFRS, PPP_BFRSIZE) ;
	for (ivec = INT_FIRST; ivec <= INT_LAST; ivec++)
		{
		char far *temp = (char far *) getvect(_ppp_interrupt) ;
		if (!_fmemcmp(&temp[3], _pkt_line, strlen(_pkt_line)))
			{
			if (is_ppp_driver(_ppp_interrupt)) break ;
			}
		}

	if (_ppp_interrupt > INT_LAST)
		{
		fprintf(stderr, "No PPP packet driver found!\r\n") ;
		exit(255) ;
		}

	_ppp_interrupt = ivec ;

	regs.r_ax = PD_ACCESS | PD_PPP ;
	regs.r_bx = 0xFFFF ;            /* any type */
	regs.r_dx = 0 ;			/* if number */
	regs.r_cx = 0 ;
	regs.r_ds = FP_SEG(fptr_ppp_ip_type) ;
	regs.r_si = FP_OFF(fptr_ppp_ip_type) ;
	regs.r_es = FP_SEG(fptr_ppp_isr) ;
	regs.r_di = FP_OFF(fptr_ppp_isr) ;

	intr(_ppp_interrupt, &regs) ;
	if (regs.r_flags & CARRY)
		{
		fprintf(stderr, "ERROR # 0x%02X accessing packet driver\n",
			regs.r_dx >> 8) ;
		exit(255) ;
		}

	_ppp_handle = regs.r_ax;
	atexit(ppp_fini) ;
	}

PUBLIC byte FAR *ppp_received(void)
	{
	word oldin, newin;	/* ip sequence numbers */
	word oldest ;
	int i ;

	/* -------------------------------------------------- */
	/* Check if there are any pending and take the oldest */
	/* -------------------------------------------------- */
	oldest = oldin = 0xFFFF ;
	for (i = 0; i < PPP_MAXBFRS; i++)
		{
		if (_ppp_pkts[i][0] != 1) continue ;

	        newin = *((word FAR *) &_ppp_pkts[i][6]) ;
		if (newin <= oldin)
			{
			oldin = newin ;
			oldest = i ;
			}
		}

	return (oldest == 0xFFFF) ? NULL : &_ppp_pkts[oldest][2] ;
	}

PUBLIC int ppp_send(byte FAR *buffer, unsigned length)
	{
	struct REGPACK regs ;
	void FAR *fptr ;
	int retries ;

	fptr = (void FAR *) buffer ;
	for (retries = 0; retries < 5; retries++)
		{
		regs.r_ax = PD_SEND ;
	        regs.r_ds = FP_SEG(fptr) ;
		regs.r_si = FP_OFF(fptr) ;
	        regs.r_cx = length ;
		intr(_ppp_interrupt, &regs) ;
		if (regs.r_flags & CARRY) continue ;
		return 0 ;
		}

	return 1 ;
	}

PUBLIC void ppp_free(byte FAR *pkt)
	{
	pkt[-2] = 0 ;
	}

PRIVATE void ppp_fini(void)
	{
	struct REGPACK regs ;

	regs.r_ax = PD_RELEASE ;
	regs.r_bx = _ppp_handle ;
	intr(_ppp_interrupt, &regs) ;
	if (regs.r_flags & CARRY)
		{
		fprintf(stderr, "ERROR releasing PPP packet driver!\n");
		}
	}

PRIVATE int is_ppp_driver(int ivec)
	{
	static word _ppp_ip_type = 0x0008 ;	/* intelled value */
	static void far *fptr_ppp_ip_type = (void far *) &_ppp_ip_type ;
	static void far *fptr_ppp_isr = (void far *) _ppp_isr ;
	struct REGPACK regs ;
	int pd_type ;

	/* lets find out about the driver */
	regs.r_ax = PD_DRIVER_INFO ;
	intr(_ppp_interrupt, &regs) ;

	if (regs.r_flags & CARRY)
		{
		/* handle old versions, assume a class and just keep trying */
		for (pd_type = 1; pd_type < 128; ++pd_type)
			{
			regs.r_ax = PD_ACCESS | PD_PPP ;
			regs.r_bx = pd_type ;
			regs.r_dx = 0 ;
			regs.r_cx = 0 ;
			regs.r_ds = FP_SEG(fptr_ppp_ip_type) ;
			regs.r_si = FP_OFF(fptr_ppp_ip_type) ;
			regs.r_es = FP_SEG(fptr_ppp_isr) ;
			regs.r_di = FP_OFF(fptr_ppp_isr) ;
			intr(_ppp_interrupt, &regs) ;
			if (regs.r_flags & CARRY) continue ;

			/* we found a working type, so kill it */
			ppp_fini() ;
			return 1 ;
			}

		return 0 ;
		}

	return (regs.r_cx >> 8) == PD_PPP ;
	}

