/*
	File: SOCKS4.C
*/

#include "..\include\tcp.h"
#include "..\include\socks4.h"
#include <string.h>
#include <malloc.h>

#define	DEBUG	1

typedef struct SOCKET
	{
	unsigned	port ;
	long		ip ;
	} SOCKET ;

typedef struct SOCKS4_REQUEST
	{
	byte		version ;
	byte		command ;
	SOCKET		dst ;
	char		user_id[1] ;
	} SOCKS4_REQUEST ;

typedef struct SOCKS4_REPLY
	{
	byte		version ;
	byte		result ;
	SOCKET		dst ;
	} SOCKS4_REPLY ;

PRIVATE long		_proxy_ip			= 0L		;
PRIVATE unsigned	_proxy_port			= 1080		;
PRIVATE char *		_proxy_user_id			= ""		;
PRIVATE SOCKET		_proxy_socket			= {0, 0L}	;
PRIVATE BOOL		_initialized			= FALSE		;
PRIVATE void		(*_yield)()			= NULL		;

void socks4_init(long proxy_ip, unsigned proxy_port, char *user_id, void (*yield)())
	{
	_proxy_ip	= proxy_ip ;
	_proxy_port	= proxy_port ;
	_proxy_user_id	= user_id ;
	_yield		= yield ;
	_initialized	= TRUE ;
	}

BOOL socks4_connect(tcp_Socket *s, long dst_ip, word dst_port)
	{
	SOCKS4_REQUEST *rqst ;
	SOCKS4_REPLY reply ;
	int bytes, err ;

#ifdef	DEBUG
{
char ip_bfr[16] ;
printf("Enter socks4_connect(dst=%s:%u)\n",
inet_ntoa(ip_bfr, dst_ip), dst_port) ;
}
#endif
	if (!_initialized) return FALSE ;

	/* ---------------------------------------- */
	/* (1) Allocate memory for a request packet */
	/* ---------------------------------------- */
	bytes = sizeof(SOCKS4_REQUEST) + strlen(_proxy_user_id) ;
	rqst = (SOCKS4_REQUEST *) malloc(bytes) ;
	if (!rqst)
		{
#ifdef	DEBUG
printf("Error in socks4_connect: Can't allocate request packet.\n") ;
#endif
		return FALSE ;
		}

	/* ---------------------------------------------------- */
	/* (2) Attempt an active open to the Socks Proxy Server */
	/* ---------------------------------------------------- */
	if (!tcp_open(s, 0, _proxy_ip, _proxy_port, NULL))
		{
#ifdef	DEBUG
printf("Error in socks4_connect: tcp_open() to proxy server failed.\n") ;
#endif
		return FALSE ;
		}
	sock_wait_established(s, sock_delay, _yield, NULL) ;
	sock_mode(s, TCP_MODE_BINARY) ;

	/* ------------------------------------------------- */
	/* (3) Format and send request to Socks Proxy Server */
	/* ------------------------------------------------- */
	rqst->version  = 4 ;	/* SOCKS Version 4 */
	rqst->command  = 1 ;	/* Connect Request */
	rqst->dst.ip   = intel(dst_ip) ;
	rqst->dst.port = intel16(dst_port) ;
	strcpy(rqst->user_id, _proxy_user_id) ;
	err = sock_write(s, (byte *) rqst, bytes) != bytes ;
	free(rqst) ;
	if (err)
		{
#ifdef DEBUG
printf("Error in socks4_connect: sock_write() failed.\n") ;
#endif
		sock_abort(s) ;
		return FALSE ;
		}

	/* -------------------------------------------------------- */
	/* (4) Wait for, read, and interpret the (one) reply packet */
	/* -------------------------------------------------------- */
	bytes = sock_read(s, (byte *) &reply, sizeof(SOCKS4_REPLY)) ;
	if (bytes == sizeof(SOCKS4_REPLY) &&
	    reply.version == 0 && reply.result == 0x90)
		{
		struct sockaddr sock ;

		getsockname(s, &sock, NULL) ;
		_proxy_socket.ip   = sock.s_ip ;
		_proxy_socket.port = sock.s_port ;
#ifdef	DEBUG
printf("Exit socks4_connect - connection established!\n") ;
#endif
		return TRUE ; /* Connected and ready to go! */
		}
#ifdef	DEBUG
	else
		{
printf("socks4_connect: Bad reply (bytes=%u, vers=0x%02X, result=0x%02X)\n",
bytes, reply.version, reply.result) ;
		}
#endif

sock_err:
#ifdef	DEBUG
printf("Error in socks4_connect: sock_wait_established() failed.\n") ;
#endif
	sock_abort(s) ;
	return FALSE ;
	}

BOOL socks4_bind(tcp_Socket *s, long dst_ip, word pri_port)
	{
	SOCKS4_REQUEST *rqst ;
	SOCKS4_REPLY reply ;
	int bytes, err ;

#ifdef	DEBUG
{
char ip_bfr[16] ;
printf("Enter socks4_bind(dst=%s:%u)\n",
inet_ntoa(ip_bfr, dst_ip), pri_port) ;
}
#endif
	if (!_initialized) return FALSE ;

	/* ---------------------------------------- */
	/* (1) Allocate memory for a request packet */
	/* ---------------------------------------- */
	bytes = sizeof(SOCKS4_REQUEST) + strlen(_proxy_user_id) ;
	rqst = (SOCKS4_REQUEST *) malloc(bytes) ;
	if (!rqst)
		{
#ifdef	DEBUG
printf("Error in socks4_bind: Can't allocate request packet.\n") ;
#endif
		return FALSE ;
		}

	/* ---------------------------------------------------- */
	/* (2) Attempt an active open to the Socks Proxy Server */
	/* ---------------------------------------------------- */
	if (!tcp_open(s, 0, _proxy_ip, _proxy_port, NULL))
		{
#ifdef	DEBUG
printf("Error in socks4_bind: tcp_open() to proxy server failed.\n") ;
#endif
		return FALSE ;
		}
	sock_wait_established(s, sock_delay, _yield, NULL) ;
	sock_mode(s, TCP_MODE_BINARY) ;

	/* ------------------------------------------------- */
	/* (3) Format and send request to Socks Proxy Server */
	/* ------------------------------------------------- */
	rqst->version  = 4 ;	/* SOCKS Version 4 */
	rqst->command  = 2 ;	/* Bind Request */
	rqst->dst.ip   = intel(dst_ip) ;
	rqst->dst.port = intel16(pri_port) ;
	strcpy(rqst->user_id, _proxy_user_id) ;
	err = sock_write(s, (byte *) rqst, bytes) != bytes ;
	free(rqst) ;
	if (err)
		{
#ifdef DEBUG
printf("Error in socks4_bind: sock_write() failed.\n") ;
#endif
		sock_abort(s) ;
		return FALSE ;
		}

	/* --------------------------------------------------------------- */
	/* (4) Wait for, read, and interpret the (1st of two) reply packet */
	/* --------------------------------------------------------------- */
	bytes = sock_read(s, (byte *) &reply, sizeof(SOCKS4_REPLY)) ;
	if (bytes == sizeof(SOCKS4_REPLY) &&
	    reply.version == 0 && reply.result == 0x90)
		{
		reply.dst.ip   = intel(reply.dst.ip) ;
		reply.dst.port = intel16(reply.dst.port) ;
		if (!reply.dst.ip) reply.dst.ip = _proxy_ip ;
		_proxy_socket = reply.dst ;
#ifdef	DEBUG
{
char ip_bfr[16] ;
printf("Exit socks4_bind - proxy listening at %s:%u!\n",
inet_ntoa(ip_bfr, _proxy_socket.ip), _proxy_socket.port) ;
}
#endif
		return TRUE ; /* Success: Proxy server waiting for connect */
		}
#ifdef	DEBUG
	else
		{
printf("socks4_bind: Bad reply (bytes=%u, vers=0x%02X, result=0x%02X)\n",
bytes, reply.version, reply.result) ;
		}
#endif

sock_err:
#ifdef	DEBUG
printf("Error in socks4_bind: sock_wait_established() failed.\n") ;
#endif
	sock_abort(s) ;
	return FALSE ;
	}

void socks4_getsockname(tcp_Socket *s, struct sockaddr *sock, int *len)
	{
	getsockname(s, sock, len) ;
	if (*len >= 4) sock->s_port = _proxy_socket.port ;
	if (*len >= 8) sock->s_ip   = _proxy_socket.ip ;
	}

BOOL socks4_listen(tcp_Socket *s, unsigned timeout)
	{
	SOCKS4_REPLY reply ;
	long timer ;
	int bytes ;

#ifdef	DEBUG
printf("Enter socks4_listen(timeout=%u)\n", timeout) ;
#endif
	timer = set_timeout(timeout) ;
	while (sock_rbleft(s) < sizeof(SOCKS4_REPLY))
		{
		if (!tcp_tick(s))
			{
#ifdef	DEBUG
printf("Error in socks4_listen: Lost connection w/proxy server\n") ;
#endif
			goto sock_err ;
			}
		if (chk_timeout(timer))
			{
#ifdef	DEBUG
printf("Error in socks4_listen: Timeout reached \n") ;
#endif
			goto sock_err ;
			}
		if (_yield) (*_yield)() ;
		}

	bytes = sock_read(s, (byte *) &reply, sizeof(SOCKS4_REPLY)) ;
	if (bytes == sizeof(SOCKS4_REPLY) &&
	    reply.version == 0 && reply.result == 0x90)
		{
#ifdef	DEBUG
printf("Success in socks4_listen: Passive connection established!\n") ;
#endif
		return TRUE ;
		}
#ifdef	DEBUG
	else
		{
printf("socks4_listen: Bad reply (bytes=%u, vers=0x%02X, result=0x%02X)\n",
bytes, reply.version, reply.result) ;
		}
#endif

sock_err:
	sock_abort(s) ;
	return FALSE ;
	}
